import json
import pathlib
import shutil
import signal

import brownie
import eth_account
import pytest
from click.testing import CliRunner

from beamer.cli import main
from beamer.util import TokenMatchChecker


def _generate_deployment_dir(output_dir, root, contracts):
    data = {
        "beamer_commit": "0" * 40,
        "L2": {
            str(brownie.chain.id): {
                "RequestManager": {
                    "address": contracts.request_manager.address,
                    "deployment_block": 1,
                },
                "FillManager": {"address": contracts.fill_manager.address, "deployment_block": 1},
                "ResolutionRegistry": {
                    "address": contracts.resolution_registry.address,
                    "deployment_block": 1,
                },
            }
        },
    }
    with output_dir.joinpath("deployment.json").open("wt") as f:
        json.dump(data, f)

    src = root / "contracts/build/contracts"
    shutil.copy(src / "RequestManager.json", output_dir)
    shutil.copy(src / "FillManager.json", output_dir)
    shutil.copy(src / "ResolutionRegistry.json", output_dir)


def test_cli(config, tmp_path, contracts):
    key = "0x3ff6c8dfd3ab60a14f2a2d4650387f71fe736b519d990073e650092faaa621fa"
    acc = eth_account.Account.from_key(key)
    obj = eth_account.account.create_keyfile_json(acc.key, b"")
    keyfile = tmp_path / f"{acc.address}.json"
    keyfile.write_text(json.dumps(obj))
    root = pathlib.Path(__file__).parents[3]
    deployment_dir = tmp_path / "deployment"
    deployment_dir.mkdir()
    _generate_deployment_dir(deployment_dir, root, contracts)

    signal.signal(signal.SIGALRM, lambda *_unused: signal.raise_signal(signal.SIGINT))
    signal.setitimer(signal.ITIMER_REAL, 2)

    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            "--keystore-file",
            str(keyfile),
            "--password",
            "",
            "--l1-rpc-url",
            config.l1_rpc_url,
            "--l2a-rpc-url",
            config.l2a_rpc_url,
            "--l2b-rpc-url",
            config.l2b_rpc_url,
            "--deployment-dir",
            str(deployment_dir),
            "--token-match-file",
            str(root / "beamer/data/tokens.example.json"),
        ],
    )
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "token_chain_ids_validity",
    [(["28", "588"], True), (["10", "42161", "288"], True), (["10", "28", "588"], False)],
)
def test_token_lists_validity(token_chain_ids_validity):
    token_address = "0x0000000000000000000000000000000000000001"
    token_chain_ids, valid = token_chain_ids_validity
    tokens = [
        # Parametrized token map
        [[chain_id, token_address] for chain_id in token_chain_ids],
        # Valid token map
        [
            ["28", "0x2De6a0f9dDFCb338AF1a126Dc77af9a245bBc83d"],
            ["588", "0xD184D3515e1817DDE870a2F30DEC29a8f1192414"],
        ],
    ]

    if not valid:
        with pytest.raises(AssertionError):
            TokenMatchChecker(tokens)
    else:
        TokenMatchChecker(tokens)



import brownie
from eth_abi.packed import encode_abi_packed
from eth_utils import keccak

from beamer.tests.util import alloc_accounts, create_request_hash


def test_fill_request(fill_manager, token, deployer):
    chain_id = brownie.web3.eth.chain_id
    amount = 100
    filler, receiver = alloc_accounts(2)

    with brownie.reverts("Ownable: caller is not the owner"):
        fill_manager.addAllowedLP(deployer, {"from": filler})

    whitelist_tx = fill_manager.addAllowedLP(deployer, {"from": deployer})
    assert "LPAdded" in whitelist_tx.events

    token.approve(fill_manager.address, amount, {"from": deployer})
    fill_manager.fillRequest(
        1,
        chain_id,
        token.address,
        receiver,
        amount,
        {"from": deployer},
    )

    with brownie.reverts("Already filled"):
        fill_manager.fillRequest(
            1,
            chain_id,
            token.address,
            receiver,
            amount,
            {"from": deployer},
        )

    fill_manager.removeAllowedLP(deployer, {"from": deployer})
    with brownie.reverts("Sender not whitelisted"):
        fill_manager.fillRequest(
            1,
            chain_id,
            token.address,
            receiver,
            amount,
            {"from": deployer},
        )

    blacklist_tx = fill_manager.removeAllowedLP(deployer, {"from": deployer})
    assert "LPRemoved" in blacklist_tx.events


def test_invalidate_valid_fill_hash(fill_manager, token, deployer):
    chain_id = brownie.web3.eth.chain_id
    amount = 100
    (receiver,) = alloc_accounts(1)

    fill_manager.addAllowedLP(deployer, {"from": deployer})

    token.approve(fill_manager.address, amount, {"from": deployer})
    request_id = 1
    tx = fill_manager.fillRequest(
        request_id,
        chain_id,
        token.address,
        receiver,
        amount,
        {"from": deployer},
    )
    fill_id = tx.return_value
    request_hash = create_request_hash(
        request_id,
        chain_id,
        chain_id,
        token.address,
        receiver.address,
        amount,
    )
    with brownie.reverts("Fill hash valid"):
        fill_manager.invalidateFillHash(request_hash, fill_id, chain_id)


def test_invalidated_fill_hash_event(fill_manager):
    request_hash = "1234" + "00" * 30
    fill_id = "5678" + "00" * 30
    chain_id = brownie.web3.eth.chain_id

    tx = fill_manager.invalidateFillHash(request_hash, fill_id, chain_id)

    fill_hash = keccak(
        encode_abi_packed(
            ["bytes32", "bytes32"],
            [
                bytes.fromhex(request_hash),
                bytes.fromhex(fill_id),
            ],
        )
    )

    assert "HashInvalidated" in tx.events
    assert tx.events["HashInvalidated"]["requestHash"] == "0x" + request_hash
    assert tx.events["HashInvalidated"]["fillId"] == "0x" + fill_id
    assert tx.events["HashInvalidated"]["fillHash"] == "0x" + fill_hash.hex()



import signal
import sys
from pathlib import Path
from typing import Optional

import click
import structlog

import beamer.contracts
import beamer.util
from beamer.agent import Agent
from beamer.config import Config
from beamer.l1_resolution import get_relayer_executable
from beamer.typing import URL
from beamer.util import account_from_keyfile

log = structlog.get_logger(__name__)


def _sigint_handler(agent: Agent) -> None:
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    log.info("Received SIGINT, shutting down")
    agent.stop()


@click.command()
@click.option(
    "--keystore-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
    required=True,
    metavar="FILE",
    help="The file that stores the key for the account to be used.",
)
@click.password_option(required=True, help="The password needed to unlock the account.")
@click.option(
    "--l1-rpc-url",
    type=str,
    required=True,
    metavar="URL",
    help="The URL of the L1 chain RPC server (e.g. http://10.0.0.3:8545).",
)
@click.option(
    "--l2a-rpc-url",
    type=str,
    required=True,
    metavar="URL",
    help="The URL of the source L2 chain RPC server (e.g. http://10.0.0.2:8545).",
)
@click.option(
    "--l2b-rpc-url",
    type=str,
    required=True,
    metavar="URL",
    help="The URL of the target L2 chain RPC server (e.g. http://10.0.0.3:8545).",
)
@click.option(
    "--deployment-dir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    required=True,
    metavar="DIR",
    help="The directory containing contract deployment files.",
)
@click.option(
    "--token-match-file",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=Path),
    required=True,
    metavar="FILE",
    help="The file containing token matching information.",
)
@click.option(
    "--fill-wait-time",
    type=int,
    default=120,
    help="Time in seconds to wait for a fill event before challenging a false claim.",
)
@click.option(
    "--log-level",
    type=click.Choice(("debug", "info", "warning", "error", "critical")),
    default="info",
    show_default=True,
    help="The log level.",
)
@click.option(
    "--prometheus-metrics-port",
    type=int,
    default=None,
    metavar="PORT",
    show_default=True,
    help="Provide Prometheus metrics on PORT.",
)
@click.version_option()
def main(
    keystore_file: Path,
    password: str,
    l1_rpc_url: URL,
    l2a_rpc_url: URL,
    l2b_rpc_url: URL,
    deployment_dir: Path,
    token_match_file: Path,
    fill_wait_time: int,
    log_level: str,
    prometheus_metrics_port: Optional[int],
) -> None:
    beamer.util.setup_logging(log_level=log_level.upper(), log_json=False)

    if get_relayer_executable() is None:
        log.error("No relayer found")
        sys.exit(1)

    account = account_from_keyfile(keystore_file, password)
    log.info(f"Using account {account.address}")
    deployment_info = beamer.contracts.load_deployment_info(deployment_dir)
    config = Config(
        account=account,
        deployment_info=deployment_info,
        l1_rpc_url=l1_rpc_url,
        l2a_rpc_url=l2a_rpc_url,
        l2b_rpc_url=l2b_rpc_url,
        token_match_file=token_match_file,
        fill_wait_time=fill_wait_time,
        prometheus_metrics_port=prometheus_metrics_port,
    )

    signal.signal(signal.SIGINT, lambda *_unused: _sigint_handler(agent))
    agent = Agent(config)
    agent.start()
    agent.wait()



from copy import deepcopy
from unittest.mock import MagicMock, patch

import pytest
from hexbytes import HexBytes
from web3.types import ChecksumAddress, Wei

from beamer.chain import claim_request, process_claims, process_requests
from beamer.events import (
    ClaimMade,
    InitiateL1ResolutionEvent,
    RequestCreated,
    RequestFilled,
    RequestResolved,
)
from beamer.state_machine import process_event
from beamer.tests.agent.unit.utils import (
    ACCOUNT,
    ADDRESS1,
    BLOCK_NUMBER,
    CLAIM_ID,
    CLAIMER_STAKE,
    NULL_ADDRESS,
    REQUEST_ID,
    SOURCE_CHAIN_ID,
    TARGET_CHAIN_ID,
    TERMINATION,
    TIMESTAMP,
    make_claim_challenged,
    make_claim_unchallenged,
    make_context,
    make_request,
)
from beamer.tests.agent.utils import make_address, make_tx_hash
from beamer.tests.constants import FILL_ID
from beamer.typing import FillId, Termination, TokenAmount


def test_skip_not_self_filled():
    context, _ = make_context()
    request = make_request()

    context.requests.add(request.id, request)

    assert request.is_pending  # pylint:disable=no-member
    claim_request(request, context)
    assert request.is_pending  # pylint:disable=no-member


def test_ignore_expired():
    context, config = make_context()

    request = make_request()
    request.filler = config.account.address
    context.requests.add(request.id, request)

    assert request.is_pending  # pylint:disable=no-member
    claim_request(request, context)
    assert request.is_ignored  # pylint:disable=no-member


def test_request_garbage_collection_without_claim():
    context, config = make_context()

    request = make_request()
    request.fill(config.account.address, b"", b"", TIMESTAMP)
    request.withdraw()

    context.requests.add(request.id, request)

    assert len(context.requests) == 1
    process_requests(context)
    assert len(context.requests) == 0

    request = make_request()
    request.ignore()
    context.requests.add(request.id, request)

    assert len(context.requests) == 1
    process_requests(context)
    assert len(context.requests) == 0


def test_request_garbage_collection_with_claim():
    context, config = make_context()

    request = make_request()
    request.fill(config.account.address, b"", b"", TIMESTAMP)
    request.withdraw()

    claim = make_claim_challenged(request)

    context.requests.add(request.id, request)
    context.claims.add(claim.id, claim)

    assert len(context.requests) == 1
    assert len(context.claims) == 1

    # While the claim exists, the request is not removed
    process_claims(context)
    process_requests(context)
    assert len(context.requests) == 1
    assert len(context.claims) == 1

    claim.withdraw()

    # Once the claim is removed, the request is removed as well
    process_claims(context)
    process_requests(context)
    assert len(context.requests) == 0
    assert len(context.claims) == 0


def test_handle_request_resolved():
    context, config = make_context()
    filler = make_address()
    fill_id = FILL_ID

    # Must store the result in the request
    request = make_request()
    request.fill(config.account.address, b"", fill_id, TIMESTAMP)
    request.try_to_claim()

    event = RequestResolved(
        chain_id=TARGET_CHAIN_ID,
        tx_hash=HexBytes(""),
        request_hash=request.request_hash,
        filler=filler,
        fill_id=fill_id,
        block_number=BLOCK_NUMBER,
    )

    # Without a request, we simply drop the event
    assert process_event(event, context) == (True, None)

    # Adding the request and claim to context
    context.requests.add(request.id, request)
    claim = make_claim_unchallenged(request, fill_id=fill_id)
    context.claims.add(claim.id, claim)

    assert request.l1_resolution_filler is None
    assert process_event(event, context) == (True, None)
    assert request.l1_resolution_filler == filler


def test_handle_generate_l1_resolution_event():
    context, config = make_context()

    request = make_request()
    request.fill(config.account.address, b"", b"", TIMESTAMP)
    context.requests.add(request.id, request)

    claim = make_claim_challenged(
        request=request,
        claimer=config.account.address,
        challenger=make_address(),
        challenger_stake=Wei(CLAIMER_STAKE + 1),
    )
    context.claims.add(claim.id, claim)

    event = deepcopy(claim.latest_claim_made)
    flag, events = process_event(event, context)

    assert flag
    assert events == [
        InitiateL1ResolutionEvent(
            chain_id=TARGET_CHAIN_ID,
            request_id=REQUEST_ID,
            claim_id=CLAIM_ID,
        )
    ]


def test_maybe_claim_no_l1():
    context, config = make_context()

    request = make_request()
    request.fill(config.account.address, b"", b"", TIMESTAMP)
    context.requests.add(request.id, request)

    # Claimer doesn't win challenge game, `withdraw` must not be called
    claim = make_claim_challenged(
        request=request,
        claimer=config.account.address,
        claimer_stake=Wei(10),
        challenger_stake=Wei(50),
    )
    context.claims.add(claim.id, claim)

    # Make sure we're outside the challenge period
    block = context.latest_blocks[request.source_chain_id]
    assert block["timestamp"] >= claim.termination

    assert not claim.transaction_pending
    process_claims(context)
    assert not claim.transaction_pending

    # Claimer wins challenge game, `withdraw` must be called
    claim = make_claim_challenged(
        request=request,
        claimer=config.account.address,
        claimer_stake=Wei(100),
        challenger_stake=Wei(50),
    )
    context.claims.add(claim.id, claim)

    # Make sure we're outside the challenge period
    block = context.latest_blocks[request.source_chain_id]
    assert block["timestamp"] >= claim.termination

    assert not claim.transaction_pending
    process_claims(context)
    assert claim.transaction_pending

    # Claimer leads challenge game, but challenge period is not over
    block = context.latest_blocks[request.source_chain_id]
    claim = make_claim_challenged(
        request=request,
        claimer=config.account.address,
        claimer_stake=Wei(100),
        challenger_stake=Wei(50),
        termination=Termination(TIMESTAMP + 1),
    )
    context.claims.add(claim.id, claim)

    # Make sure we're inside the challenge period
    assert block["timestamp"] < claim.termination

    assert not claim.transaction_pending
    process_claims(context)
    assert not claim.transaction_pending


@patch("beamer.chain._withdraw")
@pytest.mark.parametrize("termination", [TIMESTAMP - 1, TIMESTAMP])
@pytest.mark.parametrize("l1_filler", [ACCOUNT.address, make_address()])
@pytest.mark.parametrize("l1_fill_id", [FILL_ID, FillId(b"wrong fill id")])
def test_maybe_claim_l1_as_claimer(
    mocked_withdraw, termination: Termination, l1_filler: ChecksumAddress, l1_fill_id: FillId
):
    context, config = make_context()
    # Assert that ACCOUNT.address is the agent's address
    # Is used as an assurance that in case ACCOUNT is changed
    assert config.account.address == ACCOUNT.address

    request = make_request()
    request.fill(config.account.address, b"", b"", TIMESTAMP)
    request.l1_resolve(l1_filler, l1_fill_id)
    context.requests.add(request.id, request)

    claim = make_claim_challenged(
        request=request, claimer=config.account.address, fill_id=FILL_ID, termination=termination
    )
    context.claims.add(claim.id, claim)

    assert not claim.transaction_pending
    process_claims(context)

    # If agent is the correct filler, `withdraw` should be called, otherwise not
    if l1_filler == context.address and l1_fill_id == FILL_ID:
        assert mocked_withdraw.called
    else:
        assert not mocked_withdraw.called


@patch("beamer.chain._withdraw")
@pytest.mark.parametrize("termination", [TIMESTAMP - 1, TIMESTAMP])
@pytest.mark.parametrize("l1_filler", [ADDRESS1, make_address()])
@pytest.mark.parametrize("l1_fill_id", [FILL_ID, FillId(b"wrong fill id")])
def test_maybe_claim_l1_as_challenger(
    mocked_withdraw, termination: Termination, l1_filler: ChecksumAddress, l1_fill_id: FillId
):
    context, config = make_context()
    # Assert that ACCOUNT.address is the agent's address
    # Is used as an assurance that in case ACCOUNT is changed
    assert config.account.address == ACCOUNT.address
    request = make_request()
    request.fill(ADDRESS1, b"", b"", TIMESTAMP)
    request.l1_resolve(l1_filler, l1_fill_id)

    context.requests.add(request.id, request)

    claim = make_claim_challenged(
        request=request,
        claimer=ADDRESS1,
        challenger=config.account.address,
        fill_id=FILL_ID,
        termination=termination,
    )
    context.claims.add(claim.id, claim)

    assert not claim.transaction_pending
    process_claims(context)

    # If agent is the challenger and the claimer cheated,
    # `withdraw` should be called, otherwise not
    if l1_filler != ADDRESS1 or l1_fill_id != FILL_ID:
        assert mocked_withdraw.called
    else:
        assert not mocked_withdraw.called


@patch("beamer.metrics")
def test_handling_claim_during_sync(_mocked_metrics):
    """
    Tests the problem encountered in https://github.com/beamer-bridge/beamer/issues/782
    """
    context, config = make_context()

    source_token = make_address()
    target_token = make_address()
    filler = make_address()
    amount = TokenAmount(123456)

    request_event = RequestCreated(
        chain_id=SOURCE_CHAIN_ID,
        block_number=BLOCK_NUMBER,
        tx_hash=make_tx_hash(),
        request_id=REQUEST_ID,
        target_chain_id=TARGET_CHAIN_ID,
        source_token_address=source_token,
        target_token_address=target_token,
        target_address=make_address(),
        amount=amount,
        valid_until=TERMINATION,
    )
    claim_event = ClaimMade(
        chain_id=SOURCE_CHAIN_ID,
        block_number=BLOCK_NUMBER,
        tx_hash=make_tx_hash(),
        claim_id=CLAIM_ID,
        request_id=REQUEST_ID,
        fill_id=FILL_ID,
        claimer=config.account.address,
        claimer_stake=Wei(1_000),
        last_challenger=NULL_ADDRESS,
        challenger_stake_total=Wei(0),
        termination=TERMINATION,
    )
    fill_event = RequestFilled(
        chain_id=TARGET_CHAIN_ID,
        block_number=BLOCK_NUMBER,
        tx_hash=make_tx_hash(),
        request_id=REQUEST_ID,
        fill_id=FILL_ID,
        source_chain_id=SOURCE_CHAIN_ID,
        target_token_address=target_token,
        filler=filler,
        amount=amount,
    )

    context.match_checker = MagicMock()
    context.match_checker.is_valid_pair.return_value = True

    process_event(request_event, context)
    process_event(claim_event, context)
    process_event(fill_event, context)

    request = context.requests.get(REQUEST_ID)
    assert request is not None

    # Before the fix, the claim event didn't change the requests state. So it
    # remained in the pending state, which led to another try to claim the request
    assert request.is_claimed

    # It's also important that the fill information is still correctly attached to the
    # request, even if it's received later
    assert request.filler == fill_event.filler
    assert request.fill_tx == fill_event.tx_hash
    assert request.fill_id == fill_event.fill_id



import subprocess
import sys

from pathlib import Path
from typing import Optional
import structlog
from hexbytes import HexBytes

from beamer.typing import URL


log = structlog.get_logger(__name__)

_RELAYER_NAMES = {"linux": "relayer-node16-linux-x64", "darwin": "relayer-node16-macos-x64"}


def get_relayer_executable() -> Optional[Path]:
    """Returns the path to the relayer executable if available, otherwise `None`"""

    bin_dir = Path(__file__).parent.joinpath("data/relayers/")
    try:
        maybe_relayer = bin_dir.joinpath(_RELAYER_NAMES[sys.platform])
    except KeyError:
        log.warning("Unsupported platform", platform=sys.platform)
        return None

    if maybe_relayer.exists():
        log.debug("Found relayer executable", relayer=maybe_relayer)
        return maybe_relayer.resolve()

    log.warning("No relayer executable found")
    return None


def run_relayer_for_tx(l1_rpc: URL, l2_rpc: URL, privkey: str, tx_hash: HexBytes) -> None:
    relayer = get_relayer_executable()

    if relayer is None:
        log.error("No relayer found")
        sys.exit(1)

    subprocess.run(
        [
            str(relayer),
            "--l1rpcprovider",
            l1_rpc,
            "--l2rpcprovider",
            l2_rpc,
            "--l1wallet",
            privkey,
            "--froml2transactionhash",
            tx_hash,
        ],
        capture_output=True,
        encoding="utf-8",
        check=True,  # check throws an error right away
    )



from pathlib import Path
from typing import Tuple
from unittest.mock import MagicMock

from eth_account import Account
from eth_typing import BlockNumber, ChecksumAddress
from eth_utils import to_checksum_address
from hexbytes import HexBytes
from web3.constants import ADDRESS_ZERO
from web3.types import BlockData, Timestamp, Wei

from beamer.config import Config
from beamer.events import ClaimMade
from beamer.models.claim import Claim
from beamer.models.request import Request
from beamer.state_machine import Context
from beamer.tests.agent.utils import make_address
from beamer.tests.constants import FILL_ID
from beamer.tracker import Tracker
from beamer.typing import URL, ChainId, ClaimId, FillId, RequestId, Termination, TokenAmount
from beamer.util import TokenMatchChecker

SOURCE_CHAIN_ID = ChainId(2)
TARGET_CHAIN_ID = ChainId(3)

REQUEST_ID = RequestId(10)
CLAIM_ID = ClaimId(200)

CLAIMER_STAKE = Wei(10_000_000)
CHALLENGER_STAKE = Wei(5_000_000)
ZERO_STAKE = Wei(0)

TERMINATION = Termination(1)
TIMESTAMP = Timestamp(457)
BLOCK_NUMBER = BlockNumber(12345)

ACCOUNT = Account.from_key(0xB25C7DB31FEED9122727BF0939DC769A96564B2DE4C4726D035B36ECF1E5B364)
ADDRESS1 = make_address()
NULL_ADDRESS = to_checksum_address("0x0000000000000000000000000000000000000000")


class MockEth:
    def __init__(self, chain_id):
        self.chain_id = chain_id

    wait_for_transaction_receipt = MagicMock()


class MockWeb3:
    def __init__(self, chain_id):
        self.eth = MockEth(chain_id=chain_id)


def make_request(valid_until: int = TIMESTAMP - 1) -> Request:
    return Request(
        request_id=REQUEST_ID,
        source_chain_id=SOURCE_CHAIN_ID,
        target_chain_id=TARGET_CHAIN_ID,
        source_token_address=make_address(),
        target_token_address=make_address(),
        target_address=make_address(),
        amount=TokenAmount(123),
        valid_until=valid_until,
    )


def make_claim_unchallenged(
    request: Request,
    claim_id: ClaimId = CLAIM_ID,
    claimer: ChecksumAddress = None,
    claimer_stake: Wei = CLAIMER_STAKE,
    fill_id: FillId = FILL_ID,
    termination: Termination = TERMINATION,
    stay_in_started_state: bool = False,
) -> Claim:
    return make_claim_challenged(
        request=request,
        claim_id=claim_id,
        claimer=claimer,
        claimer_stake=claimer_stake,
        challenger=to_checksum_address(ADDRESS_ZERO),
        challenger_stake=Wei(0),
        fill_id=fill_id,
        termination=termination,
        stay_in_started_state=stay_in_started_state,
    )


def make_claim_challenged(
    request: Request,
    claim_id: ClaimId = CLAIM_ID,
    claimer: ChecksumAddress = None,
    claimer_stake: Wei = CLAIMER_STAKE,
    challenger: ChecksumAddress = None,
    challenger_stake: Wei = CHALLENGER_STAKE,
    fill_id: FillId = FILL_ID,
    termination: Termination = TERMINATION,
    stay_in_started_state: bool = False,
) -> Claim:
    claimer = claimer or make_address()
    challenger = challenger or make_address()

    claim = Claim(
        claim_made=ClaimMade(
            chain_id=request.source_chain_id,
            tx_hash=HexBytes(b""),
            claim_id=claim_id,
            request_id=request.id,
            fill_id=fill_id,
            claimer=claimer or make_address(),
            claimer_stake=claimer_stake,
            last_challenger=challenger,
            challenger_stake_total=challenger_stake,
            termination=termination,
            block_number=BLOCK_NUMBER,
        ),
        challenge_back_off_timestamp=123,
    )
    claim.add_challenger_stake(challenger, challenger_stake)

    if not stay_in_started_state:
        # In a challenged state, the claim must be in the challenge game states
        claim.start_challenge()

        if (
            claim.is_claimer_winning  # pylint:disable=no-member
            and challenger_stake > claimer_stake
        ):
            claim.challenge(claim.latest_claim_made)

    return claim


def make_context() -> Tuple[Context, Config]:
    checker = TokenMatchChecker([])
    config = Config(
        account=ACCOUNT,
        deployment_info={},
        l1_rpc_url=URL(""),
        l2a_rpc_url=URL(""),
        l2b_rpc_url=URL(""),
        token_match_file=Path(),
        fill_wait_time=1,
        prometheus_metrics_port=None,
    )

    context = Context(
        requests=Tracker(),
        claims=Tracker(),
        request_manager=MagicMock(),
        fill_manager=MagicMock(),
        match_checker=checker,
        address=config.account.address,
        latest_blocks={
            SOURCE_CHAIN_ID: BlockData(
                {
                    "number": BlockNumber(42),
                    "timestamp": TIMESTAMP,
                }
            )
        },
        config=config,
        web3_l1=MagicMock(),
        task_pool=MagicMock(),
        l1_resolutions={},
    )
    context.request_manager.functions.claimStake().call.return_value = 1  # type: ignore
    return context, config



import time
from dataclasses import dataclass
from typing import Iterable, Optional

import requests.exceptions
import structlog
from eth_abi.codec import ABICodec
from eth_utils.abi import event_abi_to_log_topic
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3
from web3.contract import Contract, get_event_data
from web3.types import ABIEvent, BlockData, ChecksumAddress, FilterParams, LogReceipt, Wei

from beamer.typing import (
    BlockNumber,
    ChainId,
    ClaimId,
    FillHash,
    FillId,
    RequestHash,
    RequestId,
    Termination,
    TokenAmount,
)


@dataclass(frozen=True)
class Event:
    chain_id: ChainId


@dataclass(frozen=True)
class LatestBlockUpdatedEvent(Event):
    block_data: BlockData


@dataclass(frozen=True)
class InitiateL1ResolutionEvent(Event):
    request_id: RequestId
    claim_id: ClaimId


@dataclass(frozen=True)
class InitiateL1InvalidationEvent(Event):
    claim_id: ClaimId


@dataclass(frozen=True)
class TxEvent(Event):
    block_number: BlockNumber
    tx_hash: HexBytes


@dataclass(frozen=True)
class FinalizationTimeUpdated(TxEvent):
    target_chain_id: ChainId
    finalization_time: int


@dataclass(frozen=True)
class RequestEvent(TxEvent):
    request_id: RequestId


@dataclass(frozen=True)
class RequestCreated(RequestEvent):
    target_chain_id: ChainId
    source_token_address: ChecksumAddress
    target_token_address: ChecksumAddress
    target_address: ChecksumAddress
    amount: TokenAmount
    valid_until: Termination


@dataclass(frozen=True)
class RequestFilled(RequestEvent):
    fill_id: FillId
    source_chain_id: ChainId
    target_token_address: ChecksumAddress
    filler: ChecksumAddress
    amount: TokenAmount


@dataclass(frozen=True)
class DepositWithdrawn(RequestEvent):
    receiver: ChecksumAddress


@dataclass(frozen=True)
class ClaimEvent(TxEvent):
    claim_id: ClaimId


@dataclass(frozen=True)
class ClaimMade(ClaimEvent):
    request_id: RequestId
    fill_id: FillId
    claimer: ChecksumAddress
    claimer_stake: Wei
    last_challenger: ChecksumAddress
    challenger_stake_total: Wei
    termination: Termination


@dataclass(frozen=True)
class ClaimWithdrawn(ClaimEvent):
    request_id: RequestId
    claim_receiver: ChecksumAddress


@dataclass(frozen=True)
class RequestResolved(TxEvent):
    request_hash: RequestHash
    filler: ChecksumAddress
    fill_id: FillId


@dataclass(frozen=True)
class FillHashInvalidated(TxEvent):
    fill_hash: FillHash


@dataclass(frozen=True)
class HashInvalidated(TxEvent):
    request_hash: RequestHash
    fill_id: FillId
    fill_hash: FillHash


def _camel_to_snake(s: str) -> str:
    return "".join("_" + c.lower() if c.isupper() else c for c in s).lstrip("_")


_EVENT_TYPES = dict(
    RequestCreated=RequestCreated,
    RequestFilled=RequestFilled,
    DepositWithdrawn=DepositWithdrawn,
    ClaimMade=ClaimMade,
    ClaimWithdrawn=ClaimWithdrawn,
    RequestResolved=RequestResolved,
    FillHashInvalidated=FillHashInvalidated,
    HashInvalidated=HashInvalidated,
    FinalizationTimeUpdated=FinalizationTimeUpdated,
)


def _make_topics_abi_mapping_for_contracts(contracts: Iterable[Contract]) -> dict[bytes, ABIEvent]:
    result = {}
    for contract in contracts:
        result.update(_make_topics_to_abi(contract))

    return result


def _make_topics_to_abi(contract: Contract) -> dict[bytes, ABIEvent]:
    event_abis = {}
    for abi in contract.abi:
        if abi["type"] == "event":
            event_abis[event_abi_to_log_topic(abi)] = abi  # type: ignore
    return event_abis


def _decode_event(
    codec: ABICodec, log_entry: LogReceipt, chain_id: ChainId, event_abis: dict[bytes, ABIEvent]
) -> Optional[Event]:
    topic = log_entry["topics"][0]
    event_abi = event_abis[topic]
    data = get_event_data(abi_codec=codec, event_abi=event_abi, log_entry=log_entry)
    if data.event in _EVENT_TYPES:
        kwargs = {_camel_to_snake(name): value for name, value in data.args.items()}
        kwargs["chain_id"] = chain_id
        kwargs["block_number"] = log_entry["blockNumber"]
        kwargs["tx_hash"] = log_entry["transactionHash"]
        return _EVENT_TYPES[data.event](**kwargs)
    return None


def _decode_events(
    logs: list[LogReceipt], codec: ABICodec, chain_id: ChainId, event_abis: dict[bytes, ABIEvent]
) -> list[Event]:
    events = []
    for entry in logs:
        event = _decode_event(codec, entry, chain_id, event_abis)
        if event is not None:
            events.append(event)
    return events


class EventFetcher:
    _DEFAULT_BLOCKS = 1_000
    _MIN_BLOCKS = 2
    _MAX_BLOCKS = 100_000
    _ETH_GET_LOGS_THRESHOLD_FAST = 2
    _ETH_GET_LOGS_THRESHOLD_SLOW = 5

    def __init__(self, web3: Web3, contracts: tuple[Contract, ...], start_block: BlockNumber):
        self._web3 = web3
        self._chain_id = ChainId(web3.eth.chain_id)
        self._contract_addresses = [c.address for c in contracts]
        self._next_block_number = start_block
        self._blocks_to_fetch = EventFetcher._DEFAULT_BLOCKS
        self._event_abis = _make_topics_abi_mapping_for_contracts(contracts)
        self._log = structlog.get_logger(type(self).__name__)

        for contract in contracts:
            assert (
                self._chain_id == contract.web3.eth.chain_id
            ), f"Chain id mismatch for {contract}"

    def _fetch_range(
        self, from_block: BlockNumber, to_block: BlockNumber
    ) -> Optional[list[Event]]:
        """Returns a list of events that happened in the period [from_block, to_block],
        or None if a timeout occurs."""
        self._log.debug(
            "Fetching events",
            chain_id=self._chain_id,
            contracts=self._contract_addresses,
            from_block=from_block,
            to_block=to_block,
        )
        try:
            before_query = time.monotonic()
            params: FilterParams = dict(
                fromBlock=from_block, toBlock=to_block, address=self._contract_addresses
            )
            logs = self._web3.eth.get_logs(params)
            after_query = time.monotonic()

        # Boba limits the range to 5000 blocks
        # 'ValueError: {'code': -32000, 'message': 'exceed maximum block range: 5000'}'
        except (requests.exceptions.ReadTimeout, ValueError):
            old = self._blocks_to_fetch
            self._blocks_to_fetch = max(EventFetcher._MIN_BLOCKS, old // 5)
            self._log.debug(
                "Failed to get events in time, reducing number of blocks",
                chain_id=self._chain_id,
                old=old,
                new=self._blocks_to_fetch,
            )
            return None

        except requests.exceptions.ConnectionError as exc:
            assert isinstance(self._web3.provider, HTTPProvider)
            url = self._web3.provider.endpoint_uri
            self._log.error("Connection error", url=url, exc=exc)
            # Propagate the exception upwards so we don't make further attempts.
            raise exc

        else:
            duration = after_query - before_query
            if duration < EventFetcher._ETH_GET_LOGS_THRESHOLD_FAST:
                self._blocks_to_fetch = min(EventFetcher._MAX_BLOCKS, self._blocks_to_fetch * 2)
            elif duration > EventFetcher._ETH_GET_LOGS_THRESHOLD_SLOW:
                self._blocks_to_fetch = max(EventFetcher._MIN_BLOCKS, self._blocks_to_fetch // 2)

            return _decode_events(
                logs=logs,
                codec=self._web3.codec,
                chain_id=self._chain_id,
                event_abis=self._event_abis,
            )

    def fetch(self) -> list[Event]:
        try:
            block_number = self._web3.eth.block_number
        except requests.exceptions.RequestException:
            return []

        if block_number < self._next_block_number:
            return []

        result = []
        from_block = self._next_block_number
        while from_block <= block_number:
            to_block = min(block_number, BlockNumber(from_block + self._blocks_to_fetch))
            try:
                events = self._fetch_range(from_block, to_block)
            except requests.exceptions.ConnectionError:
                break
            if events is not None:
                result.extend(events)
                from_block = BlockNumber(to_block + 1)

        self._next_block_number = from_block
        try:
            # Block number needs to be decremented here, because it is already incremented above
            block_data = self._web3.eth.get_block(from_block - 1)
        except requests.exceptions.RequestException:
            return result
        else:
            result.append(
                LatestBlockUpdatedEvent(
                    chain_id=self._chain_id,
                    block_data=block_data,
                )
            )
        return result



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from eth_account.signers.local import LocalAccount

from beamer.contracts import DeploymentInfo
from beamer.typing import URL


@dataclass
class Config:
    account: LocalAccount
    deployment_info: DeploymentInfo
    l1_rpc_url: URL
    l2a_rpc_url: URL
    l2b_rpc_url: URL
    token_match_file: Path
    fill_wait_time: int
    prometheus_metrics_port: Optional[int]



from typing import NewType
from eth_typing import (  # NOQA pylint:disable=unused-import
    Address,
    BlockNumber,
    ChecksumAddress,
    Hash32,
    HexAddress,
)

ChainId = NewType("ChainId", int)
ClaimId = NewType("ClaimId", int)
RequestId = NewType("RequestId", int)
FillId = NewType("FillId", bytes)
RequestHash = NewType("RequestHash", bytes)
FillHash = NewType("FillHash", bytes)
PrivateKey = NewType("PrivateKey", bytes)
TokenAmount = NewType("TokenAmount", int)
URL = NewType("URL", str)
Termination = NewType("Termination", int)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['beamer',
 'beamer.models',
 'beamer.tests',
 'beamer.tests.agent',
 'beamer.tests.agent.unit',
 'beamer.tests.contracts']

package_data = \
{'': ['*'], 'beamer': ['data/*', 'data/abi/*', 'data/relayers/*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'prometheus-client>=0.14.1,<0.15.0',
 'python-statemachine>=0.8.0,<0.9.0',
 'structlog>=21.5.0,<22.0.0',
 'web3>=5.24.0,<6.0.0']

entry_points = \
{'console_scripts': ['beamer-agent = beamer.cli:main']}

setup_kwargs = {
    'name': 'beamer-bridge',
    'version': '0.0.1',
    'description': 'Bridging rollups with L1 guaranteed security',
    'long_description': '[![Agent CI](https://github.com/beamer-bridge/beamer/actions/workflows/backend.yml/badge.svg)](https://github.com/beamer-bridge/beamer/actions/workflows/backend.yml)\n[![Frontend CI](https://github.com/beamer-bridge/beamer/actions/workflows/frontend.yml/badge.svg)](https://github.com/beamer-bridge/beamer/actions/workflows/frontend.yml)\n\n# Beamer Bridge\n*Transfer ERC20 assets directly between EVM compatible rollups - with a world class user experience*\n\nBeamer is a protocol to enable users to move tokens from one rollup to another.\nThe user requests a transfer by providing tokens on the source rollup.\nLiquidity providers then fill the request and directly send tokens to the user\non the target rollup.\n\nDocumentation: https://docs.beamerbridge.com\nTestnet frontend: https://testnet.beamerbridge.com\n\n\n## Running an agent from source\n\nPrerequisites: Python 3.9.x and Poetry\n\nClone this repository and enter the virtual environment:\n```\n    poetry shell\n```\n\nInstall the necessary dependencies:\n```\n    poetry install\n```\n\nFinally, still within the virtual environment, run:\n```\n    beamer-agent --keystore-file <keyfile> \\\n                 --password <keyfile-password> \\\n                 --l1-rpc-url <l1-rpc-url> \\\n                 --l2a-rpc-url <source-l2-rpc-url> \\\n                 --l2b-rpc-url <target-l2-rpc-url> \\\n                 --deployment-dir <contract-deployment-dir> \\\n                 --token-match-file <token-match-file>\n```\n\nFor more comprehensive documentation go to [Beamer documenation](https://docs.beamerbridge.com).\n',
    'author': 'Beamer Bridge Team',
    'author_email': 'contact@beamerbridge.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.beamerbridge.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import os
import time

import brownie
import pytest
from eth_utils import to_checksum_address

import beamer.agent
from beamer.tests.constants import FILL_ID
from beamer.tests.util import EventCollector, HTTPProxy, alloc_accounts, earnings, make_request


@pytest.fixture(scope="module", autouse=True)
def _allow_unlisted_pairs():
    old = os.environ.get("BEAMER_ALLOW_UNLISTED_PAIRS")
    os.environ["BEAMER_ALLOW_UNLISTED_PAIRS"] = "1"
    yield
    if old is None:
        del os.environ["BEAMER_ALLOW_UNLISTED_PAIRS"]
    else:
        os.environ["BEAMER_ALLOW_UNLISTED_PAIRS"] = old


# Scenario 1:
#
# Bob              Charlie
# --------------------------
# claim
#                  challenge
#
# Winner: Charlie
def test_challenge_1(request_manager, token, config):
    requester, charlie, target = alloc_accounts(3)

    agent = beamer.agent.Agent(config)
    agent.start()

    w3 = brownie.web3
    with earnings(w3, agent) as agent_earnings, earnings(w3, charlie) as charlie_earnings:
        token.approve(request_manager.address, 1, {"from": agent.address})
        make_request(request_manager, token, requester, target, 1, fee_data="standard")

        collector = EventCollector(request_manager, "ClaimMade")

        claim = collector.next_event()
        assert claim is not None

        agent.stop()
        agent.wait()

        request_manager.challengeClaim(
            claim.claimId, {"from": charlie, "value": claim.claimerStake + 1}
        )

        claim = collector.next_event()
        assert claim is not None
        brownie.chain.mine(timestamp=claim.termination)
        request_manager.withdraw(claim.claimId, {"from": charlie})

    assert charlie_earnings() == claim.claimerStake
    assert agent_earnings() == -claim.claimerStake


# Scenario 2:
#
# Bob              Charlie
# --------------------------
# claim
#                  challenge
# challenge
#
# Winner: Bob
def test_challenge_2(request_manager, token, config):
    requester, charlie, target = alloc_accounts(3)

    agent = beamer.agent.Agent(config)
    agent.start()

    w3 = brownie.web3
    with earnings(w3, agent) as agent_earnings, earnings(w3, charlie) as charlie_earnings:
        token.approve(request_manager.address, 1, {"from": agent.address})
        make_request(request_manager, token, requester, target, 1, fee_data="standard")

        collector = EventCollector(request_manager, "ClaimMade")

        claim = collector.next_event()
        assert claim is not None

        request_manager.challengeClaim(
            claim.claimId, {"from": charlie, "value": claim.claimerStake + 1}
        )

        # Charlie's claim.
        claim = collector.next_event()
        assert claim is not None
        assert claim.claimerStake < claim.challengerStakeTotal

        # Bob's claim.
        claim = collector.next_event()
        assert claim is not None
        assert claim.claimerStake > claim.challengerStakeTotal

        brownie.chain.mine(timestamp=claim.termination)
        request_manager.withdraw(claim.claimId, {"from": agent.address})

        agent.stop()
        agent.wait()

    assert agent_earnings() == claim.challengerStakeTotal
    assert charlie_earnings() == -claim.challengerStakeTotal


# Scenario 3:
#
# Bob              Charlie
# --------------------------
#                  claim
# challenge
#
# Winner: Bob
#
# Note: Bob is not filling the request here, merely noticing the dishonest
# claim and challenging it.
def test_challenge_3(request_manager, fill_manager, token, config):
    requester, charlie, target = alloc_accounts(3)
    agent = beamer.agent.Agent(config)

    w3 = brownie.web3
    with earnings(w3, agent) as agent_earnings, earnings(w3, charlie) as charlie_earnings:
        # Submit a request that Bob cannot fill.
        amount = token.balanceOf(agent.address) + 1
        request_id = make_request(
            request_manager, token, requester, target, amount, fee_data="standard"
        )

        stake = request_manager.claimStake()
        request_manager.claimRequest(request_id, FILL_ID, {"from": charlie, "value": stake})

        collector = EventCollector(request_manager, "ClaimMade")
        collector.next_event()

        agent.start()

        # Get Bob's challenge.
        claim = collector.next_event()
        assert claim is not None
        assert (
            claim.challengerStakeTotal > claim.claimerStake
            and claim.lastChallenger == agent.address
        )

        # Ensure that Bob did not fill the request.
        assert EventCollector(fill_manager, "RequestFilled").next_event(wait_time=2) is None

        brownie.chain.mine(timestamp=claim.termination)
        request_manager.withdraw(claim.claimId, {"from": agent.address})

        agent.stop()
        agent.wait()

    assert agent_earnings() == claim.claimerStake
    assert charlie_earnings() == -claim.claimerStake


# Scenario 4:
#
# Bob              Charlie
# --------------------------
#                  claim
# challenge
#                  challenge
#
# Winner: Charlie
#
# Note: Bob is not filling the request here, merely noticing the dishonest
# claim and challenging it.
def test_challenge_4(request_manager, fill_manager, token, config):
    requester, charlie, target = alloc_accounts(3)

    agent = beamer.agent.Agent(config)

    w3 = brownie.web3
    with earnings(w3, agent) as agent_earnings, earnings(w3, charlie) as charlie_earnings:
        # Submit a request that Bob cannot fill.
        amount = token.balanceOf(agent.address) + 1
        request_id = make_request(
            request_manager, token, requester, target, amount, fee_data="standard"
        )

        stake = request_manager.claimStake()
        request_manager.claimRequest(request_id, FILL_ID, {"from": charlie, "value": stake})

        collector = EventCollector(request_manager, "ClaimMade")
        claim = collector.next_event()

        agent.start()

        # Get Bob's challenge.
        claim = collector.next_event()
        assert claim is not None
        assert (
            claim.challengerStakeTotal > claim.claimerStake
            and claim.lastChallenger == agent.address
        )

        # Ensure that Bob did not fill the request.
        assert EventCollector(fill_manager, "RequestFilled").next_event(wait_time=2) is None

        agent.stop()
        agent.wait()

        request_manager.challengeClaim(
            claim.claimId, {"from": charlie, "value": claim.challengerStakeTotal + 1}
        )

        claim = collector.next_event()
        assert claim is not None
        assert claim.claimerStake > claim.challengerStakeTotal and claim.claimer == charlie.address

        brownie.chain.mine(timestamp=claim.termination)
        request_manager.withdraw(claim.claimId, {"from": charlie})

    assert agent_earnings() == -claim.challengerStakeTotal
    assert charlie_earnings() == claim.challengerStakeTotal


# Scenario 5:
#
# Bob              Charlie
# --------------------------
#                  fill (if honest)
#                  claim
#
# ....fill_wait_time....
#
# challenge (if not honest)
#
#
# Note: This test tests if Bob waits `fill_wait_time` seconds before challenging
# a dishonest claim
@pytest.mark.parametrize("honest_claim", [True, False])
def test_challenge_5(request_manager, fill_manager, token, config, honest_claim):
    requester, charlie, target = alloc_accounts(3)

    proxy_l2b = HTTPProxy(config.l2b_rpc_url)
    proxy_l2b.delay_rpc({"eth_getLogs": 3})
    proxy_l2b.start()

    l2b_rpc_url = "http://%s:%s" % (
        proxy_l2b.server_address[0],
        proxy_l2b.server_address[1],
    )

    config.l2b_rpc_url = l2b_rpc_url
    config.fill_wait_time = 6

    agent = beamer.agent.Agent(config)
    agent.start()

    fill_manager.addAllowedLP(charlie.address)

    # Submit a request that Bob cannot fill.
    amount = token.balanceOf(agent.address) + 1
    request_id = make_request(
        request_manager, token, requester, target, amount, fee_data="standard"
    )
    fill_id = FILL_ID

    if honest_claim:
        # Fill by Charlie
        token.mint(charlie.address, amount, {"from": charlie.address})
        token.approve(fill_manager, amount, {"from": charlie.address})
        fill_transaction = fill_manager.fillRequest(
            request_id,
            brownie.chain.id,
            token.address,
            target,
            amount,
            {"from": charlie.address},
        )
        fill_id = fill_transaction.return_value

    # claim by Charlie
    stake = request_manager.claimStake()
    request_manager.claimRequest(request_id, fill_id, {"from": charlie.address, "value": stake})

    collector = EventCollector(request_manager, "ClaimMade")
    claim = collector.next_event()
    assert claim is not None

    # Wait just before the challenge back off time
    time.sleep(config.fill_wait_time - 1)

    # Regardless of the honesty of the claim there should be no challenge event
    claim = collector.next_event(0.1)
    assert claim is None

    claim = collector.next_event()
    if honest_claim:
        # No challenge received
        assert claim is None
    else:
        # Challenge expected
        assert claim is not None
        assert claim.lastChallenger == to_checksum_address(agent.address)

    agent.stop()
    proxy_l2b.stop()
    agent.wait()


# Scenario 6:
#
# Charlie          Dave
# --------------------------
# claim
#                  challenge
#
# Winner: Charlie
#
# Note: Bob is not participating in the challenge here. We test whether Bob
# will attempt to withdraw the stakes in place of Dave.
def test_withdraw_not_participant(request_manager, token, config):
    requester, charlie, dave, target = alloc_accounts(4)

    agent = beamer.agent.Agent(config)

    # Submit a request that Bob cannot fill.
    amount = token.balanceOf(agent.address) + 1
    request_id = make_request(
        request_manager, token, requester, target, amount, fee_data="standard"
    )

    stake = request_manager.claimStake()
    request_manager.claimRequest(request_id, FILL_ID, {"from": charlie, "value": stake})

    collector = EventCollector(request_manager, "ClaimMade")
    claim = collector.next_event()
    assert claim is not None

    request_manager.challengeClaim(claim.claimId, {"from": dave, "value": stake + 1})
    brownie.chain.mine(timestamp=claim.termination)

    agent.start()

    # We wait enough time for agent to potentially withdraw
    collector = EventCollector(request_manager, "ClaimWithdrawn")
    withdraw_event = collector.next_event(wait_time=2)
    assert withdraw_event is None

    agent.stop()
    agent.wait()


# Scenario 7:
#
# Bob              Charlie              Dave
# -----------------------------------------------
#                  claim
#                                       challenge
#                  challenge
# challenge
# Winner: Bob + Dave
#
# Note: testing that the agent can handle multiparty bidding
def test_challenge_7(request_manager, fill_manager, token, config):
    requester, charlie, dave, target = alloc_accounts(4)
    agent = beamer.agent.Agent(config)

    w3 = brownie.web3
    with earnings(w3, agent) as agent_earnings, earnings(w3, dave) as dave_earnings:
        # Submit a request that Bob cannot fill.
        amount = token.balanceOf(agent.address) + 1
        request_id = make_request(request_manager, token, requester, target, amount)

        stake = request_manager.claimStake()
        request_manager.claimRequest(request_id, FILL_ID, {"from": charlie, "value": stake})

        collector = EventCollector(request_manager, "ClaimMade")
        claim = collector.next_event()
        assert claim is not None

        # Dave challenges
        request_manager.challengeClaim(
            claim.claimId, {"from": dave, "value": claim.claimerStake + 1}
        )

        claim = collector.next_event()
        assert claim is not None

        agent.start()

        # Ensure that Bob did not fill the request.
        assert EventCollector(fill_manager, "RequestFilled").next_event(wait_time=2) is None

        request_manager.challengeClaim(claim.claimId, {"from": charlie, "value": stake + 1})

        claim = collector.next_event()
        assert claim is not None
        assert claim.claimerStake > claim.challengerStakeTotal and claim.claimer == charlie.address

        # Get Bob's challenge.
        claim = collector.next_event()
        assert claim is not None
        assert (
            claim.challengerStakeTotal > claim.claimerStake
            and claim.lastChallenger == agent.address
        )

        brownie.chain.mine(timestamp=claim.termination)

        request_manager.withdraw(claim.claimId, {"from": agent.address})
        request_manager.withdraw(claim.claimId, {"from": dave})

        agent.stop()
        agent.wait()

    assert agent_earnings() == stake
    assert dave_earnings() == stake + 1
